// EFileBrowser	 (C)opyright 2006 Jonas Jarvoll
//
// This is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
// See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


#include <algorithm>

#include "searchkeeper.h"
#include "messages.h"

using namespace os;
using namespace std;

SearchKeeper :: SearchKeeper( const Messenger& cTarget ): Looper( "search_worker" )
{
	m_SearchCriteria = NULL;
	m_cTarget = cTarget;
	m_pcCurrentDir = NULL;
	m_eState = S_IDLE;
}

bool SearchKeeper :: Idle()
{
	switch( m_eState )
	{
		case S_READDIR:			// We are reading all directories
		{			
			if( m_pcCurrentDir == NULL )
			{
				m_eState = S_READY;
				return true;
			}
	                        
			if( m_cPath == "/" )
			{
				m_eState = S_READY;
		    	return true;
			} 
			else
			{
				String cName;
				
				if( m_pcCurrentDir->GetNextEntry( &cName ) == 1 )
				{
					if( !( cName == "." ) && !( cName == ".." ) )
					{
						String new_path, path;
						m_pcCurrentDir->GetPath( &path );
						new_path = path + String( "/" ) + cName;					

// printf( "%s\n", new_path.c_str() );

						if( m_SearchCriteria != NULL && m_SearchCriteria->FulfillCriteria( new_path ) )
							SendSearchResultMessage( cName );

						// If this is a folder we want to look into it unless we have already scan it
						FSNode fs( new_path );				
						if( m_SearchCriteria->GetRecursive() && fs.IsDir() && !fs.IsLink() && 
							find( m_cHistory.begin(), m_cHistory.end(), new_path ) == m_cHistory.end())
						{
							m_cHistory.push_back( path );
							m_cDirectoryStack.push( m_pcCurrentDir );
							m_pcCurrentDir = new Directory( new_path );
							m_pcCurrentDir->Rewind();
							SendCurrentSearchPath();
						}
					}

					return true;
				}
				else
				{
					if( m_cDirectoryStack.empty() )
						m_eState = S_READY;
					else
					{
						m_pcCurrentDir = m_cDirectoryStack.top();
						m_cDirectoryStack.pop();
					}

					return true;
				}
		
			}
			break;
		}
		case S_READY:		// We have read all entries and want to send a READY command
		{
			if( !m_bWaitForAddReply && !m_bWaitForPathReply )
			{
				SendSearchReadyMessage();
				m_eState = S_IDLE;

				return true;
			}
			break;
		}
	}
                
	return false;
}

bool SearchKeeper :: OkToQuit()
{
	return true;
}

void SearchKeeper :: HandleMessage( Message * pcMessage )
{
	switch( pcMessage->GetCode() )
	{
		case MSG_SEARCHKEEPER_START_SEARCH:
		{
			StartSearch( pcMessage );
			break;
		}
		case MSG_SEARCHKEEPER_ACK_SEARCH_RESULT:
		{
			m_bWaitForAddReply = false;
			if( m_cAddMap.empty() == false )
			{
				Message cMsg( MSG_SEARCHKEEPER_ADD_SEARCH_RESULT );
				int nCount = 0;

				while( m_cAddMap.empty() == false )
				{
					std::vector < FileNode >::iterator i = m_cAddMap.begin();

					FSNode cNode;

					cMsg.AddString( "name", ( *i ).m_cName );
					cMsg.AddString( "path", ( *i ).m_cPath );

					m_cAddMap.erase( i );

					if( ++nCount > 5 )
						break;	
				}

				m_bWaitForAddReply = true;
				m_cTarget.SendMessage( &cMsg );
			}
			break;
		}
		case MSG_SEARCHKEEPER_ACK_SEARCH_PATH:
		{
			m_bWaitForPathReply = false;
			if( m_cPathMap.empty() == false )
			{
				Message cMsg( MSG_SEARCHKEEPER_ACK_SEARCH_PATH );
				int nCount = 0;

				while( m_cPathMap.empty() == false )
				{
					std::vector < String >::iterator i = m_cPathMap.begin();

					cMsg.AddString( "name", ( *i ));

					m_cPathMap.erase( i );

					if( ++nCount > 5 )
						break;	
				}

				m_bWaitForPathReply = true;
				m_cTarget.SendMessage( &cMsg );
			}
			break;
		}
		case MSG_SEARCHKEEPER_STOP_SEARCH:
		{
			m_eState = S_READY;
			break;
		}
		case MSG_SEARCHKEEPER_RESTART_SEARCH:
		{
		}

		default:
			Looper::HandleMessage( pcMessage );
			break;
	}
}

void SearchKeeper :: SendSearchResultMessage( const String& cName )
{
	try
	{
		String cPath;
		m_pcCurrentDir->GetPath( &cPath );
		FileNode cFileNode( cName, cPath );
                                
		if( m_bWaitForAddReply )
		{
			m_cAddMap.push_back( cFileNode );
		}
		else
		{
			Message cMsg( MSG_SEARCHKEEPER_ADD_SEARCH_RESULT );
			cMsg.AddString( "name", cName );
			cMsg.AddString( "path", cPath );

			m_cTarget.SendMessage( &cMsg );
			m_bWaitForAddReply = true;
		}
	}
	catch( ... ) { }
}

void SearchKeeper :: SendSearchReadyMessage()
{
	Message cMsg( MSG_SEARCHKEEPER_SEARCH_READY );
	m_cTarget.SendMessage( &cMsg );
}

void SearchKeeper :: SendCurrentSearchPath()
{
	try
	{
		String cPath;
		m_pcCurrentDir->GetPath( &cPath );

		if( m_bWaitForPathReply )
		{
			m_cPathMap.push_back( cPath );
		}
		else
		{
			Message cMsg( MSG_SEARCHKEEPER_NEW_SEARCH_PATH );		
			cMsg.AddString( "path", cPath );

			m_cTarget.SendMessage( &cMsg );
			m_bWaitForPathReply = true;
		}
	}
	catch( ... ) {}
}

void SearchKeeper :: StartSearch( Message* pcMessage )
{
	// Clear all used variables
	m_cAddMap.clear();
	m_cPathMap.clear();

	while( !m_cDirectoryStack.empty() )
	{
		delete m_cDirectoryStack.top();
		m_cDirectoryStack.pop();
	}

	m_cHistory.clear();

	if( pcMessage->FindString( "path", &m_cPath.str() ) != 0 )
		return;

 	if( pcMessage->FindPointer( "searchcriteria", (void**)&m_SearchCriteria ) != 0 )
		return;

	try
	{
		m_pcCurrentDir = new Directory( m_cPath );
		m_pcCurrentDir->Rewind();	
	} 
	catch( std::exception & ) { }

	m_eState = S_READDIR;
	m_bWaitForAddReply = false;
	m_bWaitForPathReply = false;
}
