// EFileBrowser	 (C)opyright 2006 Jonas Jarvoll
//
// This is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
// See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include <util/message.h>

#include "searchtab.h"
#include "messages.h"

using namespace os;

SearchTab :: SearchTab( const Rect& cFrame, const String& cTitle, uint32 nResizeMask, uint32 nFlags ) : View( cFrame, cTitle, nResizeMask, nFlags )
{
	m_SearchKeeper = NULL;

	m_Research = new Button( Rect(), "button_research", "Refresh", new Message( MSG_SEARCHKEEPER_RESTART_SEARCH ) );
	m_Research->SetEnable( false );
	AddChild( m_Research );

	m_Stop = new Button( Rect(), "button_stop", "Stop", new Message( MSG_SEARCHKEEPER_STOP_SEARCH ) );
	AddChild( m_Stop );

	m_Label = new StringView( Rect(), "label_content", "Current path:" );
	AddChild( m_Label );

	m_CurrentPath = new TextView( Rect(), "current_path", "" );
	m_CurrentPath->SetMultiLine( false );
	m_CurrentPath->SetReadOnly( true );
	AddChild( m_CurrentPath );

	m_SearchResult = new ListView( Rect(), "search_result" );
	m_SearchResult->InsertColumn( "Name", 180 );
	m_SearchResult->InsertColumn( "Path", 100 );
	AddChild( m_SearchResult );	
}

SearchTab :: ~SearchTab()
{
	delete m_Research;
	delete m_Stop;
	delete m_Label;
	delete m_CurrentPath;
	delete m_SearchResult;
}

void SearchTab :: AttachedToWindow()
{
	// Create the search keeper
	m_SearchKeeper = new SearchKeeper( this );
	m_SearchKeeper->Run();

	View::AttachedToWindow();

	m_Research->SetTarget( this );
	m_Stop->SetTarget( this );
}

void SearchTab :: DetachedFromWindow()
{
	m_SearchKeeper->PostMessage( os::M_QUIT );
	m_SearchKeeper = NULL;
	View::DetachedFromWindow();
}

void SearchTab :: SetFrame( const Rect& cRect, bool bNotifyServer )
{
	View::SetFrame( cRect, bNotifyServer );

	Layout();
}

void SearchTab :: Paint( const Rect& cUpdateRect )
{
	// Erase background noise
	FillRect( cUpdateRect, get_default_color( COL_NORMAL ) );

	View::Paint( cUpdateRect );
}

void SearchTab :: SetSearchCriteria( String path, SearchCriteria* searchcritera )
{
	m_Path = path;
	m_SearchCriteria = (*searchcritera);
}

void SearchTab :: DoSearch()
{
	m_SearchResult->Clear();
	m_CurrentPath->SetValue( m_Path );

	Message cMsg( MSG_SEARCHKEEPER_START_SEARCH );
	cMsg.AddString( "path", m_Path );
	cMsg.AddPointer( "searchcriteria", &m_SearchCriteria );

	if( m_SearchKeeper )
	{
		m_SearchKeeper->PostMessage( &cMsg );
		m_Stop->SetEnable( true );
		m_Research->SetEnable( false );
	}
}

void SearchTab :: Layout()
{
	const float offset_x = 8.0f;
	const float offset_y = 8.0f;

	Rect cFrame = GetBounds();

	Point cResearch = m_Research->GetPreferredSize( false );
	Rect cResearchFrame( offset_x, offset_y, offset_x + cResearch.x, offset_y + cResearch.y );
	m_Research->SetFrame( cResearchFrame );

	Point cStop = m_Stop->GetPreferredSize( false );
	Rect cStopFrame( cResearchFrame.right + 10.0f, offset_y, cResearchFrame.right + 10.0f + cResearch.x, offset_y + cResearch.y );
	m_Stop->SetFrame( cStopFrame );

	Point cLabel = m_Label->GetPreferredSize( false );
	Rect cLabelFrame( offset_x, cFrame.bottom - offset_y - cLabel.y, offset_x + cLabel.x, cFrame.bottom - offset_y );
	m_Label->SetFrame( cLabelFrame );

	Point cEntry = m_CurrentPath->GetPreferredSize( false );
	Rect cEntryFrame( cLabelFrame.right + 4.0f, cFrame.bottom - offset_y - cEntry.y, cFrame.right - offset_x, cFrame.bottom - offset_y );
	m_CurrentPath->SetFrame( cEntryFrame );

	Rect cListFrame( offset_x, cResearchFrame.bottom + 5.0f, cFrame.right - offset_x, cEntryFrame.top - 5.0f );
 	m_SearchResult->SetFrame( cListFrame );
}

void SearchTab :: HandleMessage( Message* pcMessage )
{
	switch( pcMessage->GetCode() )
	{
		case MSG_SEARCHKEEPER_ADD_SEARCH_RESULT:
		{
			AddSearch( pcMessage );
			if( m_SearchKeeper )
				m_SearchKeeper->PostMessage( MSG_SEARCHKEEPER_ACK_SEARCH_RESULT, m_SearchKeeper, m_SearchKeeper );                      
			break;
		}
		case MSG_SEARCHKEEPER_SEARCH_READY:
		{
			// The search is ready
			m_Stop->SetEnable( false );
			m_Research->SetEnable( true );

			m_CurrentPath->SetValue( String("The search is ready.") );
			break;
		}
		case MSG_SEARCHKEEPER_NEW_SEARCH_PATH:
		{
			String path;
			if( pcMessage->FindString( "path", &path.str() ) == 0 )
				m_CurrentPath->SetValue( path );
			
			if( m_SearchKeeper )
				m_SearchKeeper->PostMessage( MSG_SEARCHKEEPER_ACK_SEARCH_PATH, m_SearchKeeper, m_SearchKeeper );                      
			break;
		}
		case MSG_SEARCHKEEPER_STOP_SEARCH:
		{
			if( m_SearchKeeper )
				m_SearchKeeper->PostMessage( MSG_SEARCHKEEPER_STOP_SEARCH, m_SearchKeeper, m_SearchKeeper );                      
			break;
		}
		case MSG_SEARCHKEEPER_RESTART_SEARCH:
		{
			DoSearch();
			break;
		}

		default:
			View::HandleMessage( pcMessage );
			break;
	}
}

void SearchTab :: AddSearch( Message* pcMessage )
{
	String path, name;

	if( pcMessage->FindString( "name", &name.str() ) != 0 )
		return;

	if( pcMessage->FindString( "path", &path.str() ) != 0 )
		return;

	// Add to the listview
	ListViewStringRow* pcRow = new ListViewStringRow();
	pcRow->AppendString( name );
	pcRow->AppendString( path );
	m_SearchResult->InsertRow( pcRow );
}

